/* 
   main.cpp
   Christopher Cabrera
   07/30/2013

   Code for a Q-Learning search algorithm.
 
   Reference: David Poole - Q_Learning algorithms
*/

#include <stdlib.h>
#include <time.h>
#include <vector>
#include <iostream>

const int COLS = 20;
const int ROWS = 20;

class Q_Env {
public:
    int goalX;
    int goalY;

    Q_Env( ) {
        srand(time(NULL));
        goalX = (int)(rand() % COLS);
        goalY = (int)(rand() % ROWS);
        steps = 0;
        totReward = 0.0;
	currX = (int) (rand() % COLS);
        currY = (int) (rand() % ROWS);
    }
    
    int steps;
    double totReward;
    int currX;
    int currY;
    
    double doStep( int action ) {
        int actualDirection;
        double reward;
        int newX, newY;
        int r = (int) (rand() % COLS);
        if ( r < 4 ) actualDirection = r;
        else actualDirection = action;
        
        reward = 0.0;
        
        // GOAL
        if ( currX == goalX && currY == goalY ) {
	  reward = 10.0;
            newX = ((int) (rand() % COLS));
            newY = ((int) (rand() % ROWS));
        } else {
            switch (actualDirection) {
                case 0: // UP:
                    if ( currY == 0 ) {
                        newY = currY;
                        newX = currX;
                        reward = -1;
                    } else {
                        newY = currY - 1;
                        newX = currX;
                    }
                    break;
                case 1: // RIGHT:
                    if ( currX == COLS - 1 ) {
                        newY = currY;
                        newX = currX;
                        reward = -1.0;
                    } else {
                        newY = currY;
                        newX = currX + 1;
                    }
                    break;
                case 2: // DOWN:
                    if ( currY == ROWS - 1 ) {
                        newY = currY;
                        newX = currX;
                        reward = -1.0;
                    } else {
                        newY = currY + 1;
                        newX = currX;
                    }
                    break;
                case 3: // LEFT:
                    if ( currX == 0 ) {
                        newY = currY;
                        newX = currX;
                        reward = -1.0;
                    } else {
                        newY = currY;
                        newX = currX - 1;
                    }
                    break;
            }
            }
        
        ++steps;
        totReward += reward;
        
        currX = newX;
        currY = newY;
        return reward;
        }
};

class Q_Agent {
    
public:
    Q_Env env;
    
    Q_Agent( Q_Env e )
    {
        env = e;
        discount = 0.9;
        alphaFixed = false;
        tracing = true;
        
        for ( int x = 0; x < COLS; ++x )
            for ( int y = 0; y < ROWS; ++y )
                for (int a = 0; a < 4; ++a )
                {
                    qvalues[x][y][a] = 0;
                    visits[x][y][a] = 0;
                }
    }
    
    double qvalues[COLS][ROWS][4];
    int visits[COLS][ROWS][4];
    double discount;
    bool alphaFixed;
    bool alpha;
    bool tracing;
    
    void doStep(int action, double newDiscount, double alphaFieldValue) {
        discount = newDiscount;
        alpha= alphaFieldValue;
        doStep(action);
    }
    
    void doStep(int action){
        int oldX = env.currX;
        int oldY = env.currY;
        double reward = env.doStep(action);
        
        int newX = env.currX;
        int newY = env.currY;
        
        double newVal = value(newX, newY);
        double newDatum = reward + discount * newVal;
        visits[oldX][oldY][action]++;
        if ( !alphaFixed )
            alpha = 1.0 / visits[oldX][oldY][action];
        
        if ( tracing ) {
            std::cout << "(" << oldX << ", " << oldY << ") A=" << action << " R=" << reward << " (" << newX << ", " << newY << ") newDatum=" << newDatum << "\n";
            std::cout << "     Qold = " << qvalues[oldX][oldY][action] << " VISITS=" << visits[oldX][oldY][action];
        }
        
        qvalues[oldX][oldY][action] = (1 - alpha) * qvalues[oldX][oldY][action] + alpha * newDatum;
        if ( tracing ) std::cout << "\n" << "Qnew=" << qvalues[oldX][oldY][action];
    }
    
    double value( int xVal, int yVal) {
        double val = qvalues[xVal][yVal][3];
        for (int i = 2; i >= 0; i--) {
            if ( qvalues[xVal][yVal][i] > val) {
                val = qvalues[xVal][yVal][i];
            }
        }
        return val;
    }
    
    void doSteps( int count, double greedyProb, double newDiscount, double alphaFieldValue) {
        discount = newDiscount;
        alpha = alphaFieldValue;
        for ( int i = 0; i < count; ++i )
        {
            if ( tracing )
                std::cout << "\nCOUNT= " << i + 1 << "\n";
            double r = rand();
            if ( r < greedyProb )
            {
                int startDir = (int)(rand()%4);
                double bestVal = qvalues[env.currX][env.currY][startDir];
                int bestDir = startDir;
                for ( int dir = 1; dir < 4; ++dir )
                {
                    startDir = (startDir + 1) % 4;
                    if ( qvalues[env.currX][env.currY][startDir] > bestVal ) {
                        bestVal = qvalues[env.currX][env.currY][startDir];
                        bestDir = startDir;
                    }
                }
                doStep(bestDir);
            } else {
                doStep((int) (rand() % 4));
            }
        }
    }
};

int main() {
    Q_Env env;
    Q_Agent agent( env );
    
    int steps, greedyProb = 0.75, discout = 0.9, alpha = 0.5;
    
    std::cout << "How many steps would you like to make ('q' to quit)? ";
    while ( std::cin >> steps )
    {
        agent.doSteps( steps, greedyProb, discout, alpha );
        
        std::cout << "How many steps would you like to make ('q' to quit)? ";
    }
    
}
